/*
 * Copyright (c) 2015-2019, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *  ======== event_handler.c ========
 */

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

/* TI-RTOS Kernel modules */
#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Clock.h>
#include <ti/sysbios/knl/Task.h>

/* Driver Header files */
#include <ti/drivers/GPIO.h>
#include <ti/drivers/ADC.h>

/* Driver configuration */
#include "ti_drivers_config.h"

/* Application includes */
#include "event_handler.h"
#include "uart_interconnect.h"
#include "uart_command_ids.h"

#define CLOCK_SEC_TO_TICK(x) (x * 1000 * (1000 / Clock_tickPeriod))
#define CLOCK_MSEC_TO_TICK(x) (x * (1000 / Clock_tickPeriod))

/* Event variables */
static Event_Struct eventStruct;
Event_Handle eventHandle;

/* Static shared variables */
static uint16_t lastBatteryReading = 0;

/*
 *  ======== uartIcCallback ========
 *  Incoming UART request from host
 */
static void uartIcRequestCallback(uint8_t commandId, uint8_t length, uint8_t * buffer) {
    switch (commandId) {
    case UART_IC_PING:
    {
        uartIcSendPacket(UART_IC_PONG, NULL, 0);
        break;
    }
    case UART_IC_TURN_OF_SYSTEM_REQUEST:
    {
        /* Send back a request responds */
        bool retVal = uartIcSendPacket(UART_IC_TURN_OF_SYSTEM_REQUEST_RSP, NULL, 0);

        /* If successful, turn off the system. In case this fails for some reason
         * we assume the message was lost and expect the host to re-send the request. */
        if (retVal) {
            Event_post(eventHandle, EVENT_TURN_OFF_SYSTEM);
        }
        break;
    }
    case UART_IC_GET_BATTERY_STATUS:
    {
        /* Send back the last battery level with the responds */
        uartIcSendPacket(UART_IC_GET_BATTERY_STATUS_RSP, (uint8_t *) &lastBatteryReading, 2);
        break;
    }
    default:
        break;
    }
}

/*
 *  ======== gpioCallback ========
 *  Common callback function for the GPIO interrupt.
 */
static void gpioCallback(uint_least8_t index)
{
    if (index == CONFIG_GPIO_KEY1) {
        Event_post(eventHandle, EVENT_WPS_BUTTON_PRESSED);
    } else {
        /* All other inputs will turn on the system for now  */
        Event_post(eventHandle, EVENT_TURN_ON_SYSTEM);
    }
}

/*
 *  ======== clkCallback ========
 *  Battery level sample clock callback
 */
static void clkCallback(uintptr_t arg0)
{
    /* Post event on timeout */
    Event_post(eventHandle, EVENT_UPDATE_BATTERY_LEVEL);
}

/*
 *  ======== clkShutdownCallback ========
 *  Battery level sample clock callback
 */
static void clkShutdownCallback(uintptr_t arg0)
{
    /* Post event to shutdown the device */
    Event_post(eventHandle, EVENT_TURN_OFF_SYSTEM);
}


/*
 *  ======== mainThread ========
 */
void *mainThread(void *arg0)
{
    /* Internal variables and structures */
    ADC_Params   adcParams;
    ADC_Handle   adcHandle;
    Clock_Params clkParams;
    Clock_Struct clkStruct, clkStructShutdown;
    Clock_Handle clkHandle, clkHandleShutdown;
    int_fast16_t returnValue;

    /* Internal state */
    bool systemIsActive = false;

    /* Initialize the event object */
    Event_construct(&eventStruct, NULL);
    eventHandle = Event_handle(&eventStruct);

    /* Initialize the periodic battery sample clock */
    Clock_Params_init(&clkParams);
    clkParams.period = CLOCK_SEC_TO_TICK(60*5); // 5 minute interval
    clkParams.startFlag = true;

    /* Set timeout to 0 to trigger a reading right away */
    Clock_construct(&clkStruct, (Clock_FuncPtr)clkCallback, 0, &clkParams);
    clkHandle = Clock_handle(&clkStruct);

    /* Construct the shutdown timeout clock */
    Clock_Params_init(&clkParams);
    clkParams.startFlag = false;
    Clock_construct(&clkStructShutdown, (Clock_FuncPtr)clkShutdownCallback, CLOCK_SEC_TO_TICK(30), &clkParams);
    clkHandleShutdown = Clock_handle(&clkStructShutdown);

    /* Call driver initialization functions */
    ADC_init();
    GPIO_init();

    /* Setup GPIO */
    GPIO_setConfig(CONFIG_GPIO_KEY1, GPIO_CFG_IN_PU | GPIO_CFG_IN_INT_FALLING);
    GPIO_setConfig(CONFIG_GPIO_KEY2, GPIO_CFG_IN_PU | GPIO_CFG_IN_INT_FALLING);
    GPIO_setConfig(CONFIG_GPIO_PIR_INPUT1, GPIO_CFG_IN_PD | GPIO_CFG_IN_INT_RISING);
    GPIO_setConfig(CONFIG_GPIO_PIR_INPUT2, GPIO_CFG_IN_PD | GPIO_CFG_IN_INT_RISING);

    GPIO_setCallback(CONFIG_GPIO_KEY1, gpioCallback);
    GPIO_setCallback(CONFIG_GPIO_KEY2, gpioCallback);
    GPIO_setCallback(CONFIG_GPIO_PIR_INPUT1, gpioCallback);
    GPIO_setCallback(CONFIG_GPIO_PIR_INPUT2, gpioCallback);

    /* Enable interrupts */
    GPIO_enableInt(CONFIG_GPIO_KEY1);
    GPIO_enableInt(CONFIG_GPIO_KEY2);
    GPIO_enableInt(CONFIG_GPIO_PIR_INPUT1);
    GPIO_enableInt(CONFIG_GPIO_PIR_INPUT2);

    /* Setup UART */
    returnValue = uartIcInit(115200, uartIcRequestCallback);

    /* Trap if we fail the UART init */
    while (!returnValue);

    /* Setup ADC */
    ADC_Params_init(&adcParams);
    adcHandle = ADC_open(CONFIG_ADC_0, &adcParams);

    /* Start the battery clock */
    Clock_start(clkHandle);

    /* Main event loop.
     *
     * EVENTS:
     *
     * EVENT_TURN_ON_SYSTEM - Turns on WiFi and camera domains
     * EVENT_TURN_OFF_SYSTEM - Turns of WiFi and camera domains
     * EVENT_UPDATE_BATTERY_LEVEL - Update the battery level value
     * EVENT_WPS_BUTTON_PRESSED - The WPS button was pressed
     */
    while (1) {
        /* Wait for an event */
        uint32_t event = Event_pend(eventHandle, Event_Id_NONE, EVENT_ALL, BIOS_WAIT_FOREVER);

        /* Handle each event in order*/
        while (event) {
            if (event & EVENT_TURN_ON_SYSTEM) {
                /* This only matters if the system is not already active */
                if (!systemIsActive) {
                    /* Turn on system power */
                    GPIO_write(CONFIG_GPIO_HOST_ON, 1);

                    /* TODO: Check what this means in practice but assume
                     * we first need to release the reset signal and then potentially
                     * flag the "WIFI_CONFIG" IO in some way (yet not clear what this dose)  */

                    /* Release WiFi reset after 5ms */
                    Task_sleep(CLOCK_MSEC_TO_TICK(5));
                    GPIO_write(CONFIG_GPIO_WIFI_nRESET, 1);

                    /* Configure WiFi from hibernate to active */
                    Task_sleep(CLOCK_MSEC_TO_TICK(100));
                    GPIO_write(CONFIG_GPIO_WIFI_CONFIG, 0);

                    /* System is now considered active */
                    systemIsActive = true;

                    /* Start the UART read */
                    uartIcEnableUart();

                    /* Start the shutdown timer */
                    Clock_start(clkHandleShutdown);
                }
                /* Clear event */
                event &= ~ EVENT_TURN_ON_SYSTEM;
            } else if (event & EVENT_TURN_OFF_SYSTEM) {
                /* This should only happen while the system is active but
                 * for good measure, we check it anyway to catch potential errors
                 */
                if (systemIsActive) {
                    /* Stop the shutdown timer in case it is still running */
                    Clock_stop(clkHandleShutdown);

                    /* Disable the UART to allow for standby */
                    uartIcDisableUart();

                    /* Bring-up in reverse */
                    GPIO_write(CONFIG_GPIO_WIFI_CONFIG, 1);
                    Task_sleep(CLOCK_MSEC_TO_TICK(100));
                    GPIO_write(CONFIG_GPIO_WIFI_nRESET, 0);
                    GPIO_write(CONFIG_GPIO_HOST_ON, 0);

                    /* System is now considered inactive */
                    systemIsActive = false;
                } else {
                    /* This should (likely) never happen, trap here so that we can debug */
                    while(1);
                }
                /* Clear event */
                event &= ~ EVENT_TURN_OFF_SYSTEM;
            } else if (event & EVENT_UPDATE_BATTERY_LEVEL) {
                /* Turn on the battery gauge */
                GPIO_write(CONFIG_GPIO_VBAT_MEAS, 1);

                /* Add delay */
                Task_sleep(CLOCK_MSEC_TO_TICK(1000));

                /* Sample the battery level */
                uint16_t readValue = 0;
                returnValue = ADC_convert(adcHandle, &readValue);

                if (returnValue == ADC_STATUS_SUCCESS) {
                    /* Scale the ADC value to the expected unit */
                    lastBatteryReading = 2 * readValue;
                } else {
                    /* This should never happen, trap here so that we can debug */
                    while(1);
                }

                /* Turn off the battery gauge */
                GPIO_write(CONFIG_GPIO_VBAT_MEAS, 0);

                /* Clear event */
                event &= ~ EVENT_UPDATE_BATTERY_LEVEL;
            } else if (event & EVENT_WPS_BUTTON_PRESSED) {
                bool retVal = uartIcSendPacket(UART_IC_EV_WPS_BUTTON_PRESSED, NULL, 0);

                if (!retVal) {
                    /* If it failed, sleep for a while and try again, we only retry once. */
                    Task_sleep(CLOCK_MSEC_TO_TICK(100));
                    uartIcSendPacket(UART_IC_EV_WPS_BUTTON_PRESSED, NULL, 0);
                }

                /* Clear event */
                event &= ~ EVENT_WPS_BUTTON_PRESSED;
            } else {
                /* Should never happen */
                while(1);
            }
        }
    }
}
