/*
 * Copyright (c) 2015-2019, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *  ======== uart_interconnect.c ========
 */

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

/* Driver Header files */
#include <ti/drivers/UART2.h>

/* Driver configuration */
#include "ti_drivers_config.h"

/* Application includes */
#include "uart_command_ids.h"
#include "uart_interconnect.h"
#include "event_handler.h"

#define SOF (0x55)

#define CLOCK_SEC_TO_TICK(x) ((x * 1000 * 1000)/Clock_tickPeriod)
#define CLOCK_MSEC_TO_TICK(x) ((x * 1000)/Clock_tickPeriod)

/* UART handle */
static UART2_Handle uartHandle;

/* UART buffers */
static uint8_t uartRxBuffer[4 + 255];
static uint8_t uartTxBuffer[4 + 255];

/* Internal state */
static bool uartRxActive = false;
static bool uartTxActive = false;
static uartIcCallback_t registredUserCallback = NULL;

/*
 *  ======== readCallback ========
 *  UART read callback
 */
static void readCallback(UART2_Handle handle, void *buf, size_t count, void *userArg, int_fast16_t status) {
    /* We currently do not consider partially received packets that is not complete in themselves.
     * In simpler terms this means that we expect "SOF" in byte 0 and do not handle packets
     * wrapping the buffer. We could spend more time to make this more robust and smarter but there is
     * better ways to go about this if the customer wants more complex communication. */

    if (status != UART2_STATUS_ECANCELLED) {
        uint8_t * buffer_cast = (uint8_t *) buf;

        if (buffer_cast[0] == SOF) {
            /* Check length fields for some level of sanity check */
            uint8_t length = buffer_cast[1];
            uint8_t length_inv = buffer_cast[2];

            if (length == ((~length_inv) & 0xFF)) {
                /* Everything seems to be "OK", issue user callback */
                registredUserCallback(buffer_cast[3], length, &buffer_cast[4]);
            }
        }
    }

    if (uartRxActive) {
        /* Restart the read again */
        UART2_read(uartHandle, uartRxBuffer, 259, NULL);
    }
}

/*
 *  ======== writeCallback ========
 *  UART write callback
 */
static void writeCallback(UART2_Handle handle, void *buf, size_t count, void *userArg, int_fast16_t status) {
    /* The write is done, we do not perform any checks. */
    uartTxActive = false;
}

/*
 *  ======== uartIcInit ========
 */
bool uartIcInit(uint32_t baudrate, uartIcCallback_t userCallback)
{
    /* Internal variables and structures */
    UART2_Params uartParams;

    if (userCallback) {
        /* Store callback */
        registredUserCallback = userCallback;

        /* Configure and open driver */
        UART2_Params_init(&uartParams);
        uartParams.baudRate = baudrate;
        uartParams.readReturnMode = UART2_ReadReturnMode_PARTIAL;
        uartParams.readMode = UART2_Mode_CALLBACK;
        uartParams.writeMode = UART2_Mode_CALLBACK;
        uartParams.readCallback = readCallback;
        uartParams.writeCallback = writeCallback;
        uartHandle = UART2_open(CONFIG_UART2_0, &uartParams);
    } else {
        /* A callback must be provided */
        return false;
    }

    return true;
}

/*
 *  ======== uartIcDisableUart ========
 */
void uartIcDisableUart() {
    /* To allow standby we need to cancel all ongoing UART activity.
     * First disable flags to prevent the RX to restart again once the
     * callback fire. */
    uartRxActive = false;
    uartTxActive = false;

    UART2_readCancel(uartHandle);
    UART2_writeCancel(uartHandle);
}

/*
 *  ======== uartIcEnableUart ========
 */
void uartIcEnableUart() {
    /* Start the read */
    uartRxActive = true;
    UART2_read(uartHandle, uartRxBuffer, 259, NULL);
}

/*
 *  ======== uartIcSendPacket ========
 */
bool uartIcSendPacket(uint8_t commandId, uint8_t * payload, uint8_t length) {
    bool retVal = false;

    if (!uartTxActive) {
        if ((length == 0) || ((length > 0) && (NULL != payload))) {

            /* Copy the payload to the internal buffer */
            uartTxBuffer[0] = SOF;
            uartTxBuffer[1] = length;
            uartTxBuffer[2] = ~length;
            uartTxBuffer[3] = commandId;
            if (length) {
                memcpy((void *) &uartTxBuffer[4], (void *) payload, length);
            }

            uartTxActive = true;

            uint16_t ret = UART2_write(uartHandle, uartTxBuffer, length + 4, NULL);

            if (!ret) {
                retVal = true;
            } else {
                uartTxActive = false;
            }
        }
    }

    return retVal;
}
